from ..PickList import PickList
from ..Well import TransferError
from ..tools import round_at, did_you_mean


class AssemblyPicklistGenerator:
    """Class to generate robot picklists to mix genetic parts for DNA assembly.

    (for TECAN EVO, Labcyte ECHO, etc.)
    
    Parameters
    ----------

    part_mol, part_l, part_g
      Amount of each DNA part that an assembly mix should contain (provide
      either one of these). Parameter ``part_mol`` indicates a amount of
      molecules in mol, e.g. 13e-12 to specify "13 pico-mol of each DNA part". 

    complement_to

    buffer_volume

    volume_rounding

    minimal_dispense_volume



    """

    def __init__(
        self,
        part_mol=None,
        part_l=None,
        part_g=None,
        complement_to=None,
        buffer_volume=0,
        volume_rounding=None,
        minimal_dispense_volume=0,
    ):

        self.part_mol = part_mol
        self.part_l = part_l
        self.part_g = part_g
        self.buffer_volume = buffer_volume
        self.complement_to = complement_to
        self.volume_rounding = volume_rounding
        self.minimal_dispense_volume = minimal_dispense_volume

    def make_picklist(
        self,
        assembly_plan,
        source_wells,
        destination_wells,
        buffer_well=None,
        complement_well=None,
    ):
        source_wells = list(source_wells)
        destination_wells = list(destination_wells)
        destination_wells = destination_wells[: len(assembly_plan.assemblies)]

        part_wells = {}
        duplicates = {}
        for well in source_wells:
            well_part = self.get_part_from_well(well)
            if well_part in part_wells:
                if well_part not in duplicates:
                    duplicates[well_part] = []
                duplicates[well_part].append(well.name)
                previous_well = part_wells[well_part]
                previous_concentration = previous_well.content.concentration()
                if well.content.concentration() > previous_concentration:
                    part_wells[well_part] = well
            else:
                part_wells[well_part] = well

        required_parts = set(assembly_plan.all_parts_used())
        missing_parts = required_parts.difference(set(part_wells))
        if missing_parts:
            return (
                None,
                {
                    "missing_parts": {
                        part: {
                            "featured_in": assembly_plan.assemblies_featuring(
                                part
                            ),
                            "did_you_mean": [
                                (name, str(part_wells[name]))
                                for name in did_you_mean(
                                    part, part_wells, min_score=50
                                )
                            ],
                        }
                        for part in missing_parts
                    }
                },
            )

        picklist = PickList()
        iterator = zip(assembly_plan.assemblies.items(), destination_wells)
        for ((construct_name, parts), destination_well) in iterator:
            destination_well.data.construct = construct_name
            for part in parts:
                source_well = part_wells[part]
                volume = self.volume_from_well(
                    source_well, assembly_plan.parts_data
                )
                volume = round_at(volume, self.volume_rounding)
                volume = max(volume, self.minimal_dispense_volume)
                picklist.add_transfer(
                    source_well, destination_well, volume=volume
                )

        wells_over_desired_volume = []
        if self.complement_to is not None:
            if complement_well is None:
                for label in "COMPLEMENT", "WATER":
                    complement_well = part_wells.get(label, None)
                if complement_well is None:
                    raise ValueError(
                        "Could not identify the complement well ."
                        " in the plate. One well should be marked"
                        " as containing WATER or COMPLEMENT."
                    )
            if complement_well.is_empty:
                raise TransferError(
                    "Well with COMPLEMENT/WATER is empty: %s."
                    % (complement_well)
                )
            for well in destination_wells:
                to_well = picklist.restricted_to(destination_well=well)
                total_transfer_volume = to_well.total_transfered_volume()
                complement_volume = (
                    self.complement_to
                    - total_transfer_volume
                    - self.buffer_volume
                )
                if complement_volume < 0:
                    wells_over_desired_volume.append(well)
                else:
                    picklist.add_transfer(
                        source_well=complement_well,
                        destination_well=well,
                        volume=complement_volume,
                    )
        if self.buffer_volume:
            if buffer_well is None:
                buffer_well = part_wells.get("BUFFER", None)
            if buffer_well.is_empty:
                raise TransferError("Empty buffer well: %s." % (buffer_well))
            buffer_volume = round_at(self.buffer_volume, self.volume_rounding)
            for well in destination_wells:
                picklist.add_transfer(buffer_well, well, volume=buffer_volume)

        return (
            picklist,
            {
                "wells_over_desired_volume": wells_over_desired_volume,
                "duplicates": {
                    part_name: {
                        "wells": wells,
                        "selected": part_wells[part_name],
                    }
                    for part_name, wells in duplicates.items()
                },
            },
        )

    @staticmethod
    def get_part_from_well(well):
        if "part" in well.data:
            return well.data["part"]
        else:
            return well.content.components_as_string().strip(" ")

    def get_part_molar_weight(self, part_data):
        """Returns the molar weight of the sequence in g/m.
        http://cels.uri.edu/gsc/cndna.html
        """
        if "size" in part_data:
            size = part_data["size"]
        else:
            size = len(part_data["record"])
        return 650 * size

    def volume_from_well(self, well, parts_data):
        if well.content.volume == 0:
            raise ValueError(
                ("Cannot get anything from well %s " % well)
                + "which has a volume of 0"
            )
        part_mol = self.part_mol
        part_g = self.part_g
        part_l = self.part_l
        if part_mol is not None:
            part = self.get_part_from_well(well)
            molar_weight = self.get_part_molar_weight(parts_data[part])
            part_g = molar_weight * part_mol
        if part_g is not None:
            part_l = part_g / well.content.concentration()
        return part_l



from collections import defaultdict
from ..tools import number_to_rowname, unit_factors
import pandas

def plate_to_platemap_spreadsheet(plate, wellinfo_function, filepath=None,
                                  sheet_name='plate_map', headers=True):
    """Generate a spreadsheet with a map of the plate.

    Parameters
    ----------

    plate
      A Plate object

    filepath
      Path to a ".csv" or ".xls(x)" file. Or a Pandas ExcelWriter object.

    wellinfo_function
      A function `f(well) -> info` where `well` is a Well object of the
      plate and the returned `info` is an information about the well that
      will be displayed in the well's cell in the final spreadsheet. The
      info can be a string or any other object that can be converted to a
      string

      .. code:: bash

             1  2  3  4  5  6  7  8  9  10 11 12
          A  .  .  .  .  .  .  .  .  .  .  .  .
          B  .  .  .  .  .  .  .  .  .  .  .  .
          C  .  .  .  .  .  .  .  .  .  .  .  .
          D  .  .  .  .  .  .  .  .  .  .  .  .
          E  .  .  .  .  .  .  .  .  .  .  .  .
          F  .  .  .  .  .  .  .  .  .  .  .  .
          G  .  .  .  .  .  .  .  .  .  .  .  .
          H  .  .  .  .  .  .  .  .  .  .  .  .

        """

    platedict = defaultdict(lambda *a: {})
    for well in plate:
        info = wellinfo_function(well)
        platedict[well.column][number_to_rowname(well.row)] = info
    dataframe = pandas.DataFrame.from_dict(platedict)
    if filepath is None:
        return dataframe
    elif str(filepath).lower().endswith(".csv"):
        dataframe.to_csv(filepath, header=headers, index=headers)
    else:
        dataframe.to_excel(filepath, sheet_name=sheet_name, header=headers,
                           index=headers)

def plate_to_pandas_dataframe(plate, fields=None, direction='row'):
    """Return a dataframe with the info on each well"""
    dataframe = pandas.DataFrame.from_records(plate.to_dict()["wells"]).T
    by = (["row", "column"] if direction == 'row' else ['column', 'row'])
    dataframe = dataframe.sort_values(by=by)
    if fields is not None:
        dataframe = dataframe[fields]
    return dataframe

def plate_to_content_spreadsheet(plate, filepath, content_type=None,
                                 volume_unit='uL',
                                 concentration_unit='ng-uL'):
    """Write plate into Excel with 'content', 'volume', 'concentration' sheets.

    The 'content' sheet will contain a platemap of the product contained in
    each well (e.g. a part name or a strain name)

    The 'volume' sheet will contain a platemap of the volume contained in each
    well (e.g. a part name or a strain name) in liter

    The 'concentration' sheet will contain a platemap of the concentration
    contained in each well (e.g. a part name or a strain name) in gram/liter.


    Parameters
    ----------
    plate
      Plate to be written. The content of each well must have a single product.

    filepath
      Path to the excel spreadsheet to write. An Excel writer also works.
    """

    volume_factor = unit_factors[volume_unit]
    c_mass, c_vol = concentration_unit.split('-')
    concentration_factor = unit_factors[c_mass] / unit_factors[c_vol]

    functions = [
        (('content (%s)' % content_type) if content_type else 'content',
         lambda w: w.content.components_as_string()),
        ('volume (%s)' % volume_unit,
        lambda w: w.content.volume / volume_factor),
        ('concentration (%s)' % concentration_unit,
        lambda w: w.content.concentration() / concentration_factor)
    ]
    writer = pandas.ExcelWriter(filepath)
    for name, fun in functions:
        plate_to_platemap_spreadsheet(plate, fun, sheet_name=name,
                                      filepath=writer)
    writer.close()



"""Read a .gwl picklist"""

import pandas
from ..tools import index_to_wellname
from ..PickList import PickList

def picklist_from_tecan_evo_picklist_file(filename, plates_dict):
    """Read a .gwl file into a PickList

    Parameters
    ----------

    filename
      A .gwl file

    plates_dict
      A dictionnary linking the plate names inside the file to plate objects
      For instance { "PrimersPlate": primers_plate, "SeqDestPlate": ... }

    """
    df = pandas.read_csv(filename, sep=";", header=None)
    df.columns = [
        "Action", "RackLabel", "RackID", "RackType",
        "Position", "TubeID", "Volume", "LiquidClass",
        "TipType", "TipMask"
    ]

    picklist = PickList()
    for i, row in df.iterrows():
        if row.Action == "A":
            source_plate = plates_dict[row.RackLabel]
            source_well_name = index_to_wellname(
                int(row.Position), num_wells=source_plate.num_wells,
                direction="column"
            )
            source_well = source_plate[source_well_name]
        if row.Action == "D":
            dest_plate = plates_dict[row.RackLabel]
            dest_well_name = index_to_wellname(
                int(row.Position), num_wells=dest_plate.num_wells,
                direction="column"
            )
            dest_well = dest_plate[dest_well_name]
            picklist.add_transfer(source_well, dest_well,
                                  volume=float(row.Volume))
    return picklist



import pandas as pd
from ..tools import wellname_to_index

def picklist_to_labcyte_echo_picklist_file(picklist, filename,
                                           use_well_name=True,
                                           max_dispense_volume=5e-7):
    """Write a CSV file for cherrypicking in the ECHO.

    Note that transfer volumes may have to be rounded to 2.5nl to be valid.

    Parameters
    -----------

    picklist
      The PickList object to be written in a file.

     filename
       The destination filename or path, e.g. `20160408_picklist.csv`

     use_well_name
       If True, the final file contains wellnames in plain, else the final
       file contains well numbers: A1 is 1, B1 is 2, C1 is 3, etc.

     max_dispense_volume
       Maximal volume that can be dispensed in one go. For the ECHO it may
       be 500nL (default value here). Operations above that limit will be
       decomposed into several dispensing operations, for instance 1200nL
       may be decomposed as 500+500+200nL
    """

    columns = ["Source Well", "Destination Well", "Transfer Volume"]
    picklist = picklist.enforce_maximum_dispense_volume(max_dispense_volume)

    def format_well(well_name, num_wells):
        if use_well_name:
            return well_name
        else:
            return wellname_to_index(
                well_name,
                num_wells,
                direction="column"
            )
    rows = [
        {
            "Source Well": format_well(
                transfer.source_well.name,
                transfer.source_well.plate.num_wells
            ),
            "Destination Well": format_well(
                transfer.destination_well.name,
                transfer.destination_well.plate.num_wells
            ),
            "Transfer Volume": "%.01f" % (transfer.volume / 1e-9)
        }
        for transfer in picklist.transfers_list
    ]
    df = pd.DataFrame.from_records(rows, columns=columns)
    df.to_csv(filename, sep=",", header=True, index=False)



import re
import os
from copy import deepcopy
import pandas as pd

from plateo.containers import get_plate_class
from ..tools import (
    infer_plate_size_from_wellnames,
    number_to_rowname,
    unit_factors,
)


def plate_from_dataframe(
    dataframe, wellname_field="wellname", num_wells="infer", data=None
):
    """Create a plate from a Pandas dataframe where each row contains the
    name of a well and data on the well.

    it is assumed that the dataframe's index is given by the well names.

    This function is used e.g. in `plate_from_list_spreadsheet`.

    Parameters
    ----------

    dataframe
      A Pandas dataframe

    wellname_field
      The name of the Pandas dataframe column indicating the name of the wells.

    num_wells
      Number of wells in the Plate to be created. If left to default 'infer',
      the size of the plate will be chosen as the smallest format (out of
      96, 384 and 1536 wells) which contains all the well names.

    data
      Metadata information for the plate.
    """

    # TODO: infer plate class automatically ?

    dataframe = dataframe.set_index(wellname_field)
    wells_data = {well: row.to_dict() for well, row in dataframe.iterrows()}
    if num_wells == "infer":
        num_wells = infer_plate_size_from_wellnames(wells_data.keys())
    plate_class = get_plate_class(num_wells=num_wells)
    return plate_class(wells_data=wells_data, data=data)


def plate_from_list_spreadsheet(
    filename, sheet_name=0, num_wells="infer", wellname_field="wellname"
):
    """Create a plate from a Pandas dataframe where each row contains the
    name of a well and metadata on the well.


    Parameters
    ----------

    filename
      Path to the spreadsheet file.

    sheet_name
      Index or name of the spreadsheet to use.

    num_wells
      Number of wells in the Plate to be created. If left to default 'infer',
      the size of the plate will be chosen as the smallest format (out of
      96, 384 and 1536 wells) which contains all the well names.

    wellname_field="wellname"
      Name of the column of the spreadsheet giving the well names
    """

    if ".xls" in filename:  # includes xlsx
        dataframe = pd.read_excel(filename, sheet_name=sheet_name)
    elif filename.endswith(".csv"):
        dataframe = pd.read_csv(filename)
    return plate_from_dataframe(
        dataframe,
        wellname_field=wellname_field,
        num_wells=num_wells,
        data={"filename": filename},
    )


def plate_from_platemap_spreadsheet(
    file_handle,
    file_type="auto",
    original_filename=None,
    data_field="info",
    num_wells="infer",
    plate_class=None,
    multiply_by=None,
    headers=True,
    sheet_name=0,
    skiprows=None,
):
    """Parse spreadsheets representing a plate map.

    Parameters
    ----------

    file_handle
      Either a file handle or a file path to a CSV/Excel spreadsheet. If a file
      handle is provided, then the file_type must be set, or at least the
      "original_filename".

    file_type
      Either "csv" or "excel" or "auto" (at which case the type is determined
      based on the provided file path in ``file_handle`` or
      ``original_filename``)

    original_filename
      Original filename (optional if file_handle is already a file path or
      if file_type is specified)

    data_field
      Data field of the well under which platemap's information will be stored

    num_wells
      Number of wells in the Plate to be created. If left to default 'infer',
      the size of the plate will be chosen as the smallest format (out of
      96, 384 and 1536 wells) which contains all the well names.

    headers
      Whether the spreadsheet actually writes the "A" "B", and "1" "2"


    skiprows
      Number of rows to skip (= rows before the platemap starts)

    The spreadsheet should be either a 8 rows x 12 columns csv/excel file,
    or have headers like this

    .. code:: bash

          1  2  3  4  5  6  7  8  9  10 11 12
       A  .  .  .  .  .  .  .  .  .  .  .  .
       B  .  .  .  .  .  .  .  .  .  .  .  .
       C  .  .  .  .  .  .  .  .  .  .  .  .
       D  .  .  .  .  .  .  .  .  .  .  .  .
       E  .  .  .  .  .  .  .  .  .  .  .  .
       F  .  .  .  .  .  .  .  .  .  .  .  .
       G  .  .  .  .  .  .  .  .  .  .  .  .
       H  .  .  .  .  .  .  .  .  .  .  .  .
    """
    if isinstance(file_handle, str):
        # The provided file is a file path
        original_filename = file_handle

    if file_type == "auto":
        # Determine the file type based on the file name.
        base, ext = os.path.splitext(original_filename)
        ext = ext.lower()
        if ext == ".csv":
            file_type = "csv"
        elif ext in [".xls", ".xlsx"]:
            file_type = "excel"

    index_col = 0 if headers else None
    if file_type == "csv":
        dataframe = pd.read_csv(
            file_handle,
            index_col=index_col,
            header=index_col,
            skiprows=skiprows,
        )
    elif file_type == "excel":
        dataframe = pd.read_excel(
            file_handle,
            index_col=index_col,
            sheet_name=sheet_name,
            header=index_col,
            skiprows=skiprows,
        )
    if headers:

        def compute_value(content, multiply_by, row, column):
            if multiply_by is None:
                return content
            else:
                try:
                    return content * multiply_by
                except Exception as err:
                    wellname = row + str(column)
                    raise ValueError(("In well %s: " % wellname) + str(err))
        wells_data = {
            row + str(column): {
                data_field: compute_value(content, multiply_by, row, column)
            }
            for column, column_content in dataframe.to_dict().items()
            for row, content in column_content.items()
            if isinstance(row, str)
        }
    else:
        wells_data = {
            number_to_rowname(row + 1) + str(column + 1): {data_field: content}
            for column, column_content in dataframe.to_dict().items()
            for row, content in column_content.items()
        }
    if num_wells == "infer":
        num_wells = infer_plate_size_from_wellnames(wells_data.keys())
    if plate_class is None:
        plate_class = get_plate_class(num_wells=num_wells)
    return plate_class(
        wells_data=wells_data, data={"file_source": original_filename}
    )


def plate_from_content_spreadsheet(
    spreadsheet_file,
    headers=True,
    plate_class=None,
    original_filename=None,
    sheet_names="default",
):
    """Load plate from Excel with 'content', 'volume', 'concentration' sheets.

    The 'content' sheet contains a platemap of the product contained in each
    well (e.g. a part name or a strain name)

    The 'volume' sheet contains a platemap of the volume contained in each
    well (e.g. a part name or a strain name) in liter

    The 'concentration' sheet contains a platemap of the concentration
    contained in each well (e.g. a part name or a strain name) in gram/liter.


    Parameters
    ----------
    spreadsheet_file
      Path to the excel spreadsheet, or file-like object.

    headers
      Set to True if the spreadsheets represent the platemaps using headers
      ABCDEF... for columns and 123456... for rows

    plate_class
      Class of the plate returned. If set to None, a generic Plate96 or
      Plate384 or other is returned depending on the autodetected plate size

    sheet_name
      A triple of the name of the sheets containing
    """

    # NOTE: We use deepcopy(spreadsheet_file) throughout this function because
    # Some pandas operations close file-like objects, which prevents from
    # parsing the spreadsheet several times, as this method requires.
    
    if original_filename is None:
        if isinstance(spreadsheet_file, str):
            original_filename = spreadsheet_file
        else:
            original_filename = "unknown.xlsx"
    excel = pd.ExcelFile(deepcopy(spreadsheet_file))
    sheet_names = excel.sheet_names

    sheet_name_patterns = {
        "concentration": r"[cC]oncentration(| \((\S+)-(\S+)\))",
        "volume": r"[vV]olume(| \((\S+)\))",
        "content": r"[cC]ontent(| \((\S+)\))",
    }

    default_factors = {
        "concentration": 1e-3,  # default is ng/ul = 0.001 g/l
        "volume": 1e-6,  # default is ul
        "content": "content",
    }

    field_data = {}
    for field, pattern in sheet_name_patterns.items():
        sheet_name = None
        for sheet_name in sheet_names:
            match = re.fullmatch(pattern, sheet_name)
            if match is not None:
                field_data[field] = data = {"sheet_name": sheet_name}
                groups = list(match.groups())
                if groups[0] == "":
                    data["factor"] = default_factors[field]
                elif len(groups) == 2:
                    if field == "content":
                        data["factor"] = groups[1]
                    else:
                        data["factor"] = unit_factors[groups[1]]
                elif len(groups) == 3:
                    factor = unit_factors[groups[1]] / unit_factors[groups[2]]
                    data["factor"] = factor
                else:
                    raise ValueError("Couldn't parse sheet %s" % sheet_name)
                break
        else:
            raise ValueError(
                ("No sheet found for field %s." "Check your sheet names.")
                % field
            )

    content_field_name = field_data["content"]["factor"]
    plate = plate_from_platemap_spreadsheet(
        deepcopy(spreadsheet_file),
        headers=headers,
        sheet_name=field_data["content"]["sheet_name"],
        data_field=content_field_name,
        plate_class=plate_class,
        original_filename=original_filename,
    )
    plate.merge_data_from(
        plate_from_platemap_spreadsheet(
            deepcopy(spreadsheet_file),
            data_field="volume",
            headers=headers,
            multiply_by=field_data["volume"]["factor"],
            sheet_name=field_data["volume"]["sheet_name"],
            original_filename=original_filename,
        )
    )
    plate.merge_data_from(
        plate_from_platemap_spreadsheet(
            deepcopy(spreadsheet_file),
            data_field="concentration",
            headers=headers,
            sheet_name=field_data["concentration"]["sheet_name"],
            multiply_by=field_data["concentration"]["factor"],
            original_filename=original_filename,
        )
    )

    for well in plate.iter_wells():
        content = well.data[content_field_name]
        if str(content) == "nan":
            well.data[content_field_name] = None
            continue
        volume = well.data.volume
        concentration = well.data.concentration

        try:
            well.add_content({content: volume * concentration}, volume=volume)
        except Exception as err:
            raise type(err)("Check your data for well %s" % well.name) from err
        for field in ("volume", "concentration"):
            well.data.pop(field)

    return plate



from copy import deepcopy

from .plate_to_tables import plate_to_pandas_dataframe
from ..tools import number_to_rowname


def plate_to_genesift_sequencing_order_spreadsheet(plate, output_file,
                                                   sample_name_function,
                                                   well_filter=None,
                                                   direction='row'):
    """Generate an excel spreadsheet for ordering through Genesift

    Parameters
    ----------

    plate

    output_file

    sample_name_function
      A function  f(well) => Sample name

    well_filter
      A function  f(well) => True/false

    Example:
    --------

    >>> plate_to_genesift_sequencing_order_spreadsheet(
            plate,
            output_file="genesift.xls",
            sample_name_function= lambda well: data['part_name'] ,
            well_filter=lambda well: data.get('part_name', False)
            direction='column')

    """
    if well_filter is None:
        well_filter = lambda w: True
    
    new_plate = deepcopy(plate)
    def position_function(well):
        return number_to_rowname(well.row) + "%02d" % well.column
    def sample_function(well):
        return sample_name_function(well) if well_filter(well) else None
    new_plate.compute_data_field("Position", position_function)
    new_plate.compute_data_field("Sample", sample_function)
    dataframe = plate_to_pandas_dataframe(new_plate, direction=direction)
    dataframe = dataframe[["Position", "Sample"]]
    dataframe = dataframe[[(e is not None) for e in dataframe['Sample']]]
    dataframe.to_excel(output_file, index=False)



import zipfile
import sys
PYTHON3 = (sys.version_info[0] == 3)

if PYTHON3:
    from io import StringIO, BytesIO
    StringIO = BytesIO
else:
    from StringIO import StringIO

import pandas
import matplotlib.image as mpimg
import numpy as np
from ..containers import Plate96

def plate_from_aati_fragment_analyzer_peaktable(filename):
    """"Return a Plate96 object with a data field for the ``bands``.

    Provided a ``filename`` of an AATI fragment analyzer Peak table
    (these are generally named ``{DATE} Peak Table.csv``), it generates a
    Plate96 object where each well has a data attribute "bands" of the form
    ``{peak_id: {attrs}}`` where the ``peak_id`` is a number (>1) and the attrs
    attribute has fields such as ``Size (bp)``, ``% (Conc.)``, ``nmole/L``,
    ``ng/ul``, ``RFU``.

    """
    df = pandas.read_csv(filename)
    wells = {
        name: {"bands": {
            peak_id: row.to_dict()
            for peak_id, row in d.set_index("Peak ID").iterrows()
            if row["% (Conc.)"] > 0
        }}
        for name, d in df.groupby(["Well"])
    }
    return Plate96(wells_data=wells)

def plate_from_aati_fa_gel_image(filename):
    """Return a Plate96 where each well stores an image of the gel migration

    Each well has a ``data["migration_image"]`` which is a WxH
    array, a greyscale version of the image.
    """
    img = mpimg.imread(filename)
    black_white = img.mean(axis=2)
    threshold = black_white > 0.9
    vertical_lines = (threshold.sum(axis=0) < 200).nonzero()[0]
    xmin, xmax = vertical_lines.min() + 1, vertical_lines.max() - 1
    horizontal_lines = (threshold.sum(axis=1) < 200).nonzero()[0]
    ymin, ymax = horizontal_lines.min() + 1, horizontal_lines.max() - 1
    xx = np.linspace(xmin, xmax, 97).round(0).astype(int)
    plate = Plate96("Gel Image")
    wells = plate.iter_wells(direction="column")
    bands_x = zip(xx, xx[1:])
    for (x1, x2), well in zip(bands_x, wells):
        well.data["migration_image"] = img[ymin:ymax, x1:x2]
    return plate


def plate_from_aati_fragment_analyzer_zip(filename):
    """"Return a Plate96 object with data for bands and migration image.

    Provided a zip output of an AATI fragment analyzer, it will find the
    relevant files and extract band sizes and gel images, and store these in
    each well's data.

    In the final plate, each well has a data attribute "bands" of the form
    ``{peak_id: {attrs}}`` where the ``peak_id`` is a number (>1) and the attrs
    attribute has fields such as ``Size (bp)``, ``% (Conc.)``, ``nmole/L``,
    ``ng/ul``, ``RFU``.

    Each well also has a  ``data["migration_image"]`` which is a WxH
    array, a greyscale version of the image.
    """
    ladder = None
    images_plate = None
    plate = None
    with zipfile.ZipFile(filename) as f:
        for name in f.namelist():
            if name.endswith('Peak Table.csv'):
                content = StringIO(f.read(name))
                plate = plate_from_aati_fragment_analyzer_peaktable(content)
            if name.endswith('Size Calibration.csv'):
                ladder = pandas.read_csv(StringIO(f.read(name)))
            if name.endswith(('Gel.PNG', 'Gel.JPEG')):
                content = StringIO(f.read(name))
                images_plate = plate_from_aati_fa_gel_image(content)
    if plate is None:
        raise IOError("No file `Peak Table.csv` found in AATI archive.")
    plate.data["ladder"] = ladder
    if images_plate is not None:
        plate.merge_data_from(images_plate)
    return plate



from ..containers import Plate96
import pandas

def plate_from_nanoquant_reads(nanoquant_files, plate_class=Plate96,
                               direction="column"):
    """Return a plate of DNA concentrations and qualities from NanoQuant reads.

    Parameters
    ----------

    nanoquant_files
      List of file paths or file-like objects, pointing each to a nanoquant
      read (which is a simple table of 16 concentration-quality pairs).

    plate_class
      The plateo container class of the plate to return, e.g. Plate96, or
      Plate384

    direction
      either row or column depending on which directions the wells of the
      original plates were sampled. But due to the shape of the nanoquant,
      it will certainly always be column.

    Returns
    -------

    A plate object where wells have attributes ``well.data.concentration``
    (in microg / microL) and a ``well.data.quality`` (log ratio 220 / 230).
    """
    plate = plate_class()
    wells = plate.iter_wells(direction=direction)
    for nanoquant_file in nanoquant_files:
        dataframe = pandas.read_excel(nanoquant_file, header=None)
        for i, row in dataframe.iterrows():
            well = next(wells)
            well.data["concentration"] = row[0]
            well.data["concentration_unit"] = "microg / microL"
            well.data["quality"] = row[1]
            well.data["quality_label"] = "log ratio 220/230"
    return plate



from box import Box

class TransferError(ValueError):
    pass

class WellContent:
    """Class to represent the volume and quantities of a well.

    Having the well content represented as a separate object makes it possible
    to have several wells share the same content, e.g. in throughs.
    """

    def __init__(self, quantities=None, volume=0):
        if quantities is None:
            quantities = {}
        self.volume = volume
        self.quantities = Box(quantities)

    def concentration(self, component=None, default=0):
        if self.quantities == {}:
            return default
        if self.volume == 0:
            return default
        if component is None:
            component = list(self.quantities.keys())[0]
        if component not in self.quantities:
            return default
        return 1.0 * self.quantities[component] / self.volume

    def to_dict(self):
        """Return a dict {volume: 0.0001, quantities: {...:...}}"""
        return {
            "volume": self.volume,
            "quantities": self.quantities
        }

    def make_empty(self):
        self.volume = 0
        self.quantities = Box({})

    def components_as_string(self, separator=" "):
        """Return a string representation of what's in the well mix"""
        return separator.join(sorted(self.quantities.keys()))

class Well:
    """Generic class for a well.

    Parameters
    ----------

    plate
      The plate on which the well is located

    row
      The well's row (a number, starting from 0)

    column
      The well's column (a number, starting from 0)

    name
      The well's name, for instance "A1"

    data
      A dictionnary storing data on the well, used in algorithms and reports.


    """
    capacity = None

    def __init__(self, plate, row, column, name, data=None):
        self.plate = plate
        self.row = row
        self.column = column
        self.name = name
        self.data = Box({} if data is None else data)
        self.sources = []
        self.content = WellContent()

    @property
    def volume(self):
        return self.content.volume

    def iterate_sources_tree(self):
        for source in self.sources:
            if isinstance(source, Well):
                for parent in source.iterate_sources_tree():
                    yield parent
            else:
                yield source
        yield self

    def transfer_to_other_well(self, destination_well, transfer_volume):
        if self.is_empty:
            raise TransferError(
                "Transfer %s => %s impossible: %s is empty" % (
                 self, destination_well, self))
        factor = float(transfer_volume) / self.volume

        #  pre-check in both source and destination wells that transfers
        #  are valid
        if factor > 1:
            raise TransferError(
                ("Substraction of %.2e L from %s impossible."
                 " Current volume: %.2e L")
                % (transfer_volume, self, self.volume)
            )
        final_destination_volume = destination_well.volume + transfer_volume
        if ((destination_well.capacity is not None) and
           (final_destination_volume > destination_well.capacity)):
            raise TransferError(
                "Transfer of %.2e L from %s to %s brings volume over capacity."
                % (transfer_volume, self, destination_well)
            )

        #  If you arrive here, it means that the transfer is valid, do it.
        quantities_transfered = {
            component: quantity * factor
            for component, quantity in self.content.quantities.items()
        }
        destination_well.add_content(quantities_transfered,
                                     volume=transfer_volume)
        self.subtract_content(quantities_transfered,
                              volume=transfer_volume)
        if self not in destination_well.sources:
            destination_well.sources.append(self)

    def add_content(self, components_quantities, volume=None):
        if volume > 0:
            final_volume = self.content.volume + volume
            if (self.capacity is not None) and (final_volume > self.capacity):
                raise TransferError(
                    "Transfer of %.2e L to %s brings volume over capacity."
                    % (volume, self)
                )
            self.content.volume = final_volume
        for component, quantity in components_quantities.items():
            if component not in self.content.quantities:
                self.content.quantities[component] = 0
            self.content.quantities[component] += quantity


    def subtract_content(self, components_quantities, volume=0):
        if volume > 0:
            if volume > self.volume:
                raise TransferError(
                    ("Substraction of %.2e L from %s impossible."
                     " Current volume: %.2e L")
                    % (volume, self, self.volume)
                )
            self.content.volume -= volume
        for component, quantity in components_quantities.items():
            if self.content.quantities[component] == quantity:
                self.content.quantities.pop(component)
            else:
                self.content.quantities[component] -= quantity
    
    def empty_completely(self):
        self.content.quantities = {}
        self.content.volume = 0

    @property
    def coordinates(self):
        """Return (well.row, well.column)"""
        return (self.row, self.column)

    @property
    def is_empty(self):
        """Return true iff the well's volume is 0"""
        return (self.volume == 0)


    def __repr__(self):
        return "(%s-%s)" % (self.plate.name, self.name)

    def pretty_summary(self):
        data = "\n    ".join([""] + [
            ("%s: %s" % (key, value))
            for key, value in self.data.items()])
        content = "\n    ".join([""] + [
            ("%s: %s" % (key, value))
            for key, value in self.content.quantities.items()])
        return (
            "{self}\n"
            "  Volume: {self.volume}\n"
            "  Content: {content}\n"
            "  Metadata: {data}"
        ).format(self=self, content=content, data=data)

    def to_dict(self):
        return dict(
            [
                ["name", self.name],
                ["content", self.content.to_dict()],
                ["row", self.row],
                ["column", self.column],
            ] + list(self.data.items())
        )

    def index_in_plate(self, direction='row'):
        """Return the index of the well in the plate."""
        return self.plate.wellname_to_index(self.name, direction=direction)
    
    def is_after(self, other, direction='row'):
        """Return whether this well is located strictly after the other well.
        
        Examples
        --------
        To iterate over all free wells after the last non-free well of a plate:

        >>> direction = 'row'
        >>> last_occupied_well = plate.last_nonempty_well(direction=direction)
        >>> free_wells = (w for w in plate.iter_wells(direction=direction)
        >>>               if w.is_after(last_occupied_well))
        >>> for well in free_wells: ...
        """
        well_index = self.index_in_plate(direction=direction)
        other_index = other.index_in_plate(direction=direction)
        return well_index > other_index

    def __lt__(self, other):
        return str(self) < str(other)



from ..Plate import Plate
from ..Well import Well

def get_plate_class(num_wells):
    return {
        96: Plate96,
        384: Plate384,
        1536: Plate1536,
    }[num_wells]

class Plate96(Plate):
    """Base class for standard 96-well plates"""
    num_rows = 8
    num_columns = 12


class Plate384(Plate):
    """Base class for standard 384-well plates"""
    num_rows = 16
    num_columns = 24


class Plate1536(Plate):
    """Base class for 1536-well plates (32 rows, 48 columns)"""
    num_rows = 32
    num_columns = 48

class Plate2x4(Plate):
    """Class for 8-well (2 x 4) plates such as colony plating plates"""
    num_rows = 2
    num_columns = 4


class Plate4ti0960(Plate96):
    """96-well plate from from 4titude"""
    class PlateWell(Well):
        capacity = 150e-6

class Plate4ti0130(Plate96):
    """96-well plate with 2ml deepwells from 4titude"""
    class PlateWell(Well):
        capacity = 1900e-6

class PlateLabcyteEchoLp0200Ldv(Plate384):
    """Low dead volume 384-well Echo plate"""
    class PlateWell(Well):
        capacity = 12e-6
        echo_dead_volume = 3e-6

class PlateLabcyteEchoP05525Pp(Plate384):
    """Polypropylene 384-well ECHO plate"""
    class PlateWell(Well):
        capacity = 50e-6
        echo_dead_volume = 15e-6

class Trough8x1(Plate):
    """Eight positions share the same content"""
    num_rows = 8
    num_columns = 1

    def __init__(self, name, data=None):
        Plate.__init__(self, name=name, data=data)
        for well in self:
            well.content = self["A1"].content



from collections import OrderedDict
import pandas
import flametree
from sequenticon import sequenticon_batch

from .exporters.reports import template_path, report_writer
from .tools import human_seq_size, did_you_mean


class AssemblyPlan:

    def __init__(self, assemblies, parts_data=None):
        if isinstance(assemblies, (list, tuple)):
            assemblies = OrderedDict(assemblies)
        self.assemblies = assemblies
        self.parts_data = parts_data

    def all_parts_used(self):
        return sorted(set([
            part
            for assembly in self.assemblies.values()
            for part in assembly
        ]))

    def parts_without_data(self):
        return {
            part_name: did_you_mean(part_name, self.parts_data)
            for part_name in self.all_parts_used()
            if part_name not in self.parts_data
        }

    def assemblies_featuring(self, part):
        return [name for (name, parts) in self.assemblies.items()
                if part in parts]

    def assemblies_with_records(self):
        return OrderedDict([
            (name, [self.get_part_data(p)['record'] for p in parts])
            for name, parts in self.assemblies.items()
        ])

    def get_part_data(self, part_name):
        if part_name not in self.parts_data:
            candidates = did_you_mean(part_name, self.parts_data)
            if len(candidates) == 0:
                suggestions = ""
            else:
                candidates = ", ".join(candidates)
                suggestions = "Did you mean one of those ? %s" % candidates
            raise ValueError("Unknown part %s. %s." % (part_name, suggestions))
        return self.parts_data[part_name]

    def rename_part(self, part_name, new_name):
        for name, parts in self.assemblies.items():
            for i, part in enumerate(parts):
                if part == part_name:
                    parts[i] = new_name
        if self.parts_data is not None:
            if part_name in self.parts_data:
                self.parts_data[new_name] = self.parts_data.pop(part_name)

    @staticmethod
    def from_spreadsheet(filepath=None, dataframe=None, sheet_name=0,
                         header=None):
        if dataframe is None:
            if filepath.lower().endswith('.csv'):
                with open(filepath, 'r') as f:
                    dataframe = pandas.DataFrame([
                        line.split(',')
                        for line in f.read().split('\n')
                    ])
            else:
                dataframe = pandas.read_excel(filepath, sheet_name=sheet_name,
                                              header=header)
        return AssemblyPlan(OrderedDict([
            (row[0], [
                str(e)
                for e in row[1:]
                if str(e) not in ['-', 'nan', 'None', '']
            ])
            for i, row in dataframe.iterrows()
            if str(row[0]).lower() not in ['nan', 'construct name',
                                           'construct', 'none', '']
        ]))

    def to_spreadsheet(self, path):
        with open(path, "w") as f:
            f.write("\n".join([("construct,parts")] + [
                ",".join([asm] + parts)
                for asm, parts in self.assemblies.items()
            ]))

    def assemblies_per_part(self):
        result = {}
        for assembly, parts in self.assemblies.items():
            for part in parts:
                if part not in result:
                    result[part] = set()
                result[part].add(assembly)
        for part, assemblies in result.items():
            result[part] = sorted(assemblies)
        return result

    def write_report(self, target):
        all_parts_used = self.all_parts_used()
        assemblies_per_part = self.assemblies_per_part()
        sequenticon_dict = None
        parts_length_dict = None
        first_part_data = list(self.parts_data.values())[0]
        if 'record' in first_part_data:
            sequenticons = sequenticon_batch([
                self.get_part_data(p)['record']
                for p in all_parts_used
            ], output_format='html_image')
            sequenticon_dict = OrderedDict([(name, icon)
                                            for (name, icon) in sequenticons])
        if (('record' in first_part_data) or ('size' in first_part_data)):
            parts_length_dict = {
                part: human_seq_size(self.get_part_length(part))
                for part in all_parts_used
            }
        html = report_writer.pug_to_html(
            path=template_path("assembly_plan_report.pug"),
            assembly_plan=self,
            sequenticon_dict=sequenticon_dict,
            parts_length_dict=parts_length_dict,
            all_parts_used=all_parts_used,
            assemblies_per_part=assemblies_per_part
        )
        report_writer.write_report(html, target)

    def get_part_length(self, part_name):
        data = self.get_part_data(part_name)
        if 'size' in data:
            return data['size']
        else:
            return len(data['record'])



"""Misc. file parsers that are useful for other parsers"""
from xml.sax import saxutils, parse, parseString

class ExcelHandler(saxutils.handler.ContentHandler):
    """

    This class is taken from the Python Cookbook so I guess the copyright
    goes to them.

    Memo: changed the handler from DefaultHandler to ContentHandler as the
    former doesn't seem to exist in Py2. (?)
    """

    def __init__(self):
        self.chars = []
        self.cells = []
        self.rows = []
        self.tables = []

    def characters(self, content):
        self.chars.append(content)

    def startElement(self, name, atts):
        if name == "Cell":
            self.chars = []
        elif name == "Row":
            self.cells = []
        elif name == "Table":
            self.rows = []

    def endElement(self, name):
        if name == "Cell":
            self.cells.append(''.join(self.chars))
        elif name == "Row":
            self.rows.append(self.cells)
        elif name == "Table":
            self.tables.append(self.rows)


def parse_excel_xml(xml_file=None, xml_string=None):
    """Return a list of the tables (2D arrays) in the Excel XML.

    Provide either the path to an XML file, or a string of XML content.
    """
    handler = ExcelHandler()
    if xml_file is not None:
        parse(xml_file, handler)
    elif xml_string is not None:
        parseString(xml_string, handler)
    else:
        raise ValueError("At least one of xml_file or xml_string should be"
                         " provided.")
    return handler.tables



"""Exporters to output plates or picklists in various formats."""

from .picklist_to_tecan_evo_picklist_file import \
    picklist_to_tecan_evo_picklist_file

from .picklist_to_labcyte_echo_picklist_file import \
    picklist_to_labcyte_echo_picklist_file

from .plate_to_bokeh_plots import plate_to_bokeh_plot

from .plate_to_matplotlib_plots import (PlateGraphsPlotter,
                                        PlateTextPlotter,
                                        PlateColorsPlotter)

from .plate_to_tables import (plate_to_pandas_dataframe,
                              plate_to_platemap_spreadsheet,
                              plate_to_content_spreadsheet)

from .plate_to_genesift_sequencing_order_spreadsheet import \
    plate_to_genesift_sequencing_order_spreadsheet

from .picklist_to_assembly_mix_report import picklist_to_assembly_mix_report

from .AssemblyPicklistGenerator import AssemblyPicklistGenerator



import ez_setup

ez_setup.use_setuptools()

from setuptools import setup, find_packages

exec(open("plateo/version.py").read())  # loads __version__

setup(
    name="plateo",
    version=__version__,
    author="Zulko",
    description="Read/write microplate and picklist data for lab automation",
    long_description=open("pypi-readme.rst").read(),
    license="MIT",
    keywords="microplate biology parser converter report",
    packages=find_packages(exclude="docs"),
    include_package_data=True,
    install_requires=[
        "pandas>=0.22,<1.1",
        "xlwt",
        "xlrd",
        "python-box",
        "numpy",
        "matplotlib",
        "tqdm",
        "pdf_reports",
        "flametree",
        "fuzzywuzzy",
        "sequenticon",
    ],
)



from .reports import report_writer, template_path
from ..tools import find_best_volume_unit, human_volume
from .AssemblyPicklistGenerator import AssemblyPicklistGenerator

def picklist_to_assembly_mix_report(picklist, target, data=None,
                                    volume_unit='auto'):
    html = report_writer.pug_to_html(
        path=template_path("assembly_mix_picklist_report.pug"),
        picklist=picklist,
        data=data or {},
        get_part_from_well=AssemblyPicklistGenerator.get_part_from_well,
        get_construct_from_well=lambda w: w.data.get("construct", ""),
        well_sorter=lambda w: w.coordinates[::-1],
        human_volume=human_volume
    )
    report_writer.write_report(html, target)
